// (c) Copyright 2012 Hewlett-Packard Development Company, L.P.

(function() {
"use strict";

window.Transaction = function(name, attrObj) {
	var trans = transDoc.createElement("transaction");
	if (name)
		trans.setAttribute("name", name);
	if (attrObj) {
		for (var name in attrObj) {
			trans.setAttribute(name, attrObj[name]);
		}
	}
	Transaction.init(trans);
	Transaction._transactions.appendChild(trans);
	return trans;
}

Object.extend(Transaction, {
	init : function(trans) {
		if (trans.__proto__)
			trans.__proto__ = Transaction.prototype;
		Object.extend(trans, Transaction.prototype); // Hack in FF , critical in IE
		trans.registerToDomEvents();
	},

	clear : function() {
		// TODO libs enable to clear only transactions that related to specific steps
		var trash = find("xpath:transaction", Transaction._transactions);
		for (var i = 0; i < trash.length; i++)
			trash[i].clear();

		var trans = transDoc.createElement("transactions");
		transDoc.replaceChild(trans, Transaction._transactions);
		Transaction._transactions = trans;
		stepsLUT = {};
	},

	export : function(rootStep) {
		var tempNodes = Transaction._transactions.cloneNode(true);
		// TODO libs transactions without steps are removed in persistent but stay in memory
		var vertices = find("xpath:descendant::vertex", tempNodes);
		for (var i = 0; i < vertices.length; i++) {
			var vertexStep = find(vertices[i].getAttribute("step"))[0];
			if (vertexStep && (rootStep.compareDocumentPosition(vertexStep) & Node.DOCUMENT_POSITION_CONTAINED_BY)) continue;
			var trans = vertices[i].parentNode;
			trans.removeChild(vertices[i]);
			if (find("xpath:vertex", trans).length == 0)
				trans.parentNode.removeChild(trans);
		};

		return Utils.nodeToXml(tempNodes);
	},

	import : function(transactions) {
		//Transaction.clear();
		if (!transactions)
			return;
		transactions = find("xpath:transaction", transactions);
		for (var i = 0; i < transactions.length; i++) {
			var trans = transDoc.importNode(transactions[i], true);
			if (find("xpath:transaction[@name=\"" + trans.getAttribute("name").replace(/"/g, '&quot;') + "\"]", Transaction._transactions).length > 0) {
				Log.info("Transaction with the same name already exists");
				continue;
			}

			Transaction.init(trans);

			var vertexes = find("xpath:vertex", trans);
			for (var j = 0; j < vertexes.length; j++) {
				Vertex.init(vertexes[j]);
			}

			Transaction._transactions.appendChild(trans);
		}
	},

	merge : function(transactionsNodeOrArr) {
		if (!transactionsNodeOrArr) return;
		var transactions = (Array.isArray(transactionsNodeOrArr) ? transactionsNodeOrArr : find("xpath:transaction", transactionsNodeOrArr));
		for (var i = 0, transLen = transactions.length; i < transLen; i++) {
			var newTran = transactions[i];
			var currTran = find("xpath:transaction[@name=\"" + newTran.getAttribute("name").replace(/"/g, '&quot;') + "\"]", Transaction._transactions);
			currTran = (currTran.length > 0 ? currTran[0] : null);
			if (! currTran) {
				Transaction.init(newTran);
			}

			var newVertexes = find("xpath:vertex", newTran);

			for (var j = 0, vertexLen = newVertexes.length ; j < vertexLen; j++) {
				var newVertex = newVertexes[j];
				if (currTran && find("xpath:vertex[@type=\"" + newVertex.getAttribute("type").replace(/"/g, '&quot;') +
										// Add these 2 lines once we support multiple transaction start and end points in the UI
										//  "\" and @step=\"" + newVertex.getAttribute("step").replace(/"/g, '&quot;') +
										//  "\" and @progressType=\"" + newVertex.getAttribute("progressType").replace(/"/g, '&quot;') +
										"\"]", currTran).length > 0) {
					continue;
				}
				Vertex.init(newVertex);
				if (currTran) {
					currTran.appendChild(newVertex);
				}
			}
			if (! currTran) {
				Transaction._transactions.appendChild(newTran);
			}
		}
	},

	startFromAPI :function(name, type, flags) {
		var transToStart = 0;
		var trans = find("xpath:transaction", Transaction._transactions);
		for (var i = 0; i < trans.length; i++){
			if(name == trans[i].name){
				transToStart = trans[i];
				break;
			}
		}
		if(!transToStart){
			transToStart = new Transaction(name);
			transToStart.setAttribute("needToRemove", true); // Note - the getAttribute in IE is a string in any case !!
		}
		transToStart.type = type;
		transToStart.flags = flags;
		transToStart.start();
		return;
	},

	endFromAPI   : function(name, status) {
		var trans = find("xpath:transaction", Transaction._transactions);
		for (var i = 0; i < trans.length; i++){
			if(name == trans[i].name){
				trans[i].status = status;
				trans[i].end(name);
				return;
			}
		}
	},

	hasTransactionWithinSection : function(sectionName) {
		var transArr = find('xpath:transaction', Transaction._transactions);
		for (var i = 0; i < transArr.length; ++i) {
			var vertexes = transArr[i].vertexes;
			for (var j = 0; j < vertexes.length; ++j) {
				var step = vertexes[j].step;
				if (step.containerSection === sectionName) {
					return true;
				}
			}
		}
		return false;
	},

	removeTransIfNeeded : function(name){
		var trans = find("xpath:transaction", Transaction._transactions);
		for (var i = 0; i < trans.length; i++){
			if(name == trans[i].name){
				if(trans[i].hasAttribute("needToRemove")){
					trans[i].clear();
				}
				return;
			}
		}
	},

	removeEmptyTransactions : function() {
		var emptyTransactionsFound = false ;
		var transactions = find("xpath:transaction", Transaction._transactions);
		transactions.forEach(function(transaction) {
			if (transaction.childNodes.length === 0) {
				transaction.parentNode.removeChild(transaction);
				if(!emptyTransactionsFound) emptyTransactionsFound = true;
			}
		});
		return emptyTransactionsFound;
	},

	removeVertexes : function(step) {
		var transactions = find("xpath:transaction", Transaction._transactions);
		transactions.forEach(function(transaction) {
			if (transaction) {
				transaction.removeVertex("start", step);
				transaction.removeVertex("end", step);
			}
		});
	},

	getTotalWaitTime : function(name) {
		var trans = find("xpath:transaction", Transaction._transactions);
		var totalWaitTime = 0;

		for (var i = 0; i < trans.length; i++){
			if(name == trans[i].name){
				for(var j = 0; j < Transaction.waitTimes.length; j++){
					if(Transaction.waitTimes[j].startTime >= trans[i].startTime && Transaction.waitTimes[j].duration != -1){
						totalWaitTime += Transaction.waitTimes[j].duration;
					}
				}
				break;
			}
		}
		return (totalWaitTime/1000);
	},

	getTransactionsByAttributes : function(attrObj) {
		var xpathAttrStr = "";
		for (var name in attrObj) {
			if (xpathAttrStr != "") {
				xpathAttrStr += " and ";
			}
			xpathAttrStr += "@" + name + "=\"" + attrObj[name] + "\"";
		}

		return find('xpath:transaction[' + xpathAttrStr + ']', Transaction._transactions);
	},

	registerAutoTransactions : function () {
		Query.addEventListenerUntil("TC.startRunLogic", "TC.endRunLogic", function (e) {
			function autoTransGen(name) {
				return function (e) {
					if (e.progressType == Step.PROGRESS_ACTION_STARTED) {
						// type currently 4, means automatic type; flags currently 2, means do not write to log
						Transaction.startFromAPI((typeof name == "function") ? name(e) : name, 4, 2);
						//flags currently 2, means do not write to log
					}
					if (e.progressType == Step.PROGRESS_STEP_COMPLETED) {
						Transaction.endFromAPI((typeof name == "function") ? name(e) : name, 0);
					}
				}
			}

			var mode = RTS("RTS.TruClient.AutomaticTransactions");

			if (!mode) return;

			var initHandler = autoTransGen("Transaction_Init");
			Step.sections['Init'].query.addEventListener("step/progress", initHandler);
			var actionHandler = autoTransGen("Transaction_Action");
			Step.sections['Action'].query.addEventListener("step/progress", actionHandler);
			var endHandler = autoTransGen("Transaction_End");
			Step.sections['End'].query.addEventListener("step/progress", endHandler);

			if (mode == 2) {
				var stepHandler = autoTransGen(function (e) {
					return "Transaction_" + e.target.containerSection + "_Step_" + e.target.fullIndex;
				});
				var stepQuery = Query("xpath:self::*[@level <= 3]").addEventListener("step/progress", stepHandler);
			}

			Query.addOnetimeEventListener("script/replay/end", function () {
				Step.sections['Init'].query.removeEventListener("step/progress", initHandler);
				Step.sections['Action'].query.removeEventListener("step/progress", actionHandler);
				Step.sections['End'].query.removeEventListener("step/progress", endHandler);
				if (stepQuery) stepQuery.removeEventListener("step/progress", stepHandler);
			});
		});
	}
});

Transaction.listenersAdded = 0;
Transaction.transactionCounter = 0;
Transaction.waitTimes = [];


Transaction.prototype = {
	// Properties

	get name()				{ return this.getAttribute("name");	},
	set name(value)			{ this.setAttribute("name", value);	},

	get vertexes()			{ return find("xpath:vertex", this); },

	get endVertices()		{ return find("xpath:vertex[@type='end']", this); },
	get startVertices()     { return find("xpath:vertex[@type='start']", this); },

	// Public Methods
	clear : function() {
		var verticesToClear = find("xpath:vertex", this);
		for (var i = 0; i < verticesToClear.length; i++)
			verticesToClear[i].clear();

		if (this.parentNode)
			this.parentNode.removeChild(this);
	},

	registerToDomEvents: function() {
		this.addEventListener("DOMNodeInserted", function(e) { new Event("transaction/nodeInserted", e.target, { origEvent : e }, Event.DISPATCH_SYNC);	}, false);
		this.addEventListener("DOMNodeRemoved", function(e) { new Event("transaction/nodeRemoved", e.target, { origEvent : e }, Event.DISPATCH_SYNC); }, false);
		this.addEventListener("DOMAttrModified", function(e) { new Event("transaction/attrModified" ,e.target, { origEvent : e }, Event.DISPATCH_SYNC);	}, false);
	},

	addVertex : function(typeOrVertex, step, progressType) {
		this.appendChild(typeOrVertex instanceof Vertex ? typeOrVertex : new Vertex(typeOrVertex, step, progressType));
	},

	removeVertex : function(typeOrVertex, step, progressType) {
		if (typeOrVertex instanceof Vertex) {
			if (typeOrVertex.parentNode == this) typeOrVertex.clear();
		}
		else {
			var vertexes = find('xpath:vertex[@type="'.concat(typeOrVertex, '" and @step="', step.stepId,
				progressType ? '" and @progressType="' + progressType : "",	'"]'), this);
			vertexes.forEach(function(vertex) {
				if (vertex) vertex.clear();
			});
		}
	},

	addStart	: function(step) {
		return this.addVertex("start", step, Step.PROGRESS_STEP_STARTED);
	},

	addEnd		: function(step) {
		return this.addVertex("end", step, Step.PROGRESS_AFTER_STEP_ENDED);
	},

	removeStart	: function(step, progressType) { return this.removeVertex("start", step, progressType); },
	removeEnd	: function(step, progressType) { return this.removeVertex("end", step, progressType); },


	onWaitStart : function(e){
	    Transaction.waitTimes.push({startTime: e.time.getTime(), duration : -1}) ;
	    Transaction.waitTransactionIndex = Transaction.waitTimes.length - 1;
	},

	onWaitEnd : function(e){
		var endTime = e.time.getTime();
		Transaction.waitTimes[Transaction.waitTransactionIndex].duration = endTime - Transaction.waitTimes[Transaction.waitTransactionIndex].startTime;
		Transaction.waitTransactionIndex = -1;
	},

	onPacing : function(e){
		Transaction.waitTimes.push({startTime: new Date().getTime(), duration : -1}) ;
		Query.addOnetimeEventListener("step/start step/abort", function(e) {
			if(!Transaction.transactionCounter)
				return;
			var endTime = new Date().getTime();
			if (Transaction.waitTimes.length > 0)
				Transaction.waitTimes[Transaction.waitTimes.length - 1].duration = endTime - Transaction.waitTimes[Transaction.waitTimes.length - 1].startTime;
		},false);
	},

	onStepDelay : function (e){
		Transaction.waitTimes.push({startTime: new Date().getTime(), duration : -1 }) ;
		Query.addOnetimeEventListener("step/stependdelay", function(e) {
			if(!Transaction.transactionCounter)
				return;
			var endTime = new Date().getTime();
			if (Transaction.waitTimes.length > 0)
				Transaction.waitTimes[Transaction.waitTimes.length - 1].duration = endTime - Transaction.waitTimes[Transaction.waitTimes.length - 1].startTime;
		},false);
	},

	addStepsEventListners : function(){

		if(!Transaction.listenersAdded){
			Query.addEventListener("step/wait/start", this.onWaitStart,  false);
			Query.addEventListener("step/wait/end", this.onWaitEnd, false);
			Query.addEventListener("step/start/pending", this.onPacing, false);
			Query.addEventListener("step/stepstartdelay", this.onStepDelay, false);

		}
	 	Transaction.listenersAdded = true;
	},

	RemoveStepsEventListners : function(){
		if(Transaction.listenersAdded){
			Query.removeEventListener("step/wait/start", this.onWaitStart);
			Query.removeEventListener("step/wait/end", this.onWaitEnd);
			Query.removeEventListener("step/start/pending", this.onPacing);
			Query.removeEventListener("step/stepstartdelay", this.onStepDelay);
			Transaction.waitTimes.splice(0, Transaction.waitTimes.length);
		}
		Transaction.startWaitTime = 0;
		Transaction.listenersAdded = 0;
	},


	start		: function() {
		Transaction.transactionCounter++;
		this.addStepsEventListners();
		this.startTime = new Date();
		new Event("TC.startTransaction", {}, { name: this.name }, Event.DISPATCH_SYNC);
		Event.dispatch("transaction/start", this);
	},

	end			: function() {
		this.endTime = new Date();
		var waittime = Transaction.getTotalWaitTime(this.name);
		new Event("TC.endTransaction", {}, {name: this.name, status: this.status, waitTime: waittime}, Event.DISPATCH_SYNC);
		this.type = undefined;
		Event.dispatch("transaction/end", this);
		if(Transaction.transactionCounter > 0)
			Transaction.transactionCounter--;
	    if(Transaction.transactionCounter == 0){
	        this.RemoveStepsEventListners();
	    }
	    Transaction.removeTransIfNeeded(this.name);
	 },

	isValid		: function() {
		var isEmpty = this.childNodes.length == 0;
		var properlyPaired = find('self::*[vertex[@type="start"] and vertex[@type="end"]]', this);

		return isEmpty || properlyPaired;
	}
}

function Vertex(type, step, progressType) {
	var vertex = transDoc.createElement("vertex");
	Vertex.init(vertex); //should be before settings vertex's attributes
	vertex.type = type;
	vertex.progressType = progressType;
	vertex.step = step;
	return vertex;
}

Vertex.init = function(vertex) {
	//if not IE
	if (vertex.__proto__ )
		vertex.__proto__ = Vertex.prototype;
	else
		Object.extend(vertex, Vertex.prototype);
	vertex.insertToStepsLUT();
	vertex.registerToDomEvents();
	vertex.bindInvoke();
}

Vertex.prototype = {
	// Properties
	get type()				{ return this.getAttribute("type");	},
	set type(value)			{ this.setAttribute("type", value);	},

	get step()				{ return find(this.getAttribute("step"))[0]; },
	set step(step)			{ this.setAttribute("step", step.stepId); },

	get progressType() 		{ return this.getAttribute("progressType"); },
	set progressType(value) { this.setAttribute("progressType",value); },

	// Public Methods
	insertToStepsLUT: function() {
		var step = this.step;
		if (!step) return;
		if (!(step.stepId in stepsLUT))
			stepsLUT[step.stepId] = {};
		if (!(this.progressType in stepsLUT[step.stepId]))
			stepsLUT[step.stepId][this.progressType] = [];
		stepsLUT[step.stepId][this.progressType].push(this);
	},

	registerToDomEvents: function() {
		this.addEventListener("DOMNodeInserted", function(e) { new Event("transaction/vertex/nodeInserted", e.target, { origEvent : e }, Event.DISPATCH_SYNC);	}, false);
		this.addEventListener("DOMNodeRemoved", function(e) { new Event("transaction/vertex/nodeRemoved", e.target, { origEvent : e }, Event.DISPATCH_SYNC); }, false);
		this.addEventListener("DOMAttrModified", function(e) { new Event("transaction/vertex/attrModified" ,e.target, { origEvent : e }, Event.DISPATCH_SYNC);	}, false);
	},

	clear : function() {
		if (this.parentNode) this.parentNode.removeChild(this);
		var step = this.step;
		if (step) {
			var verticesInLUT = stepsLUT[step.stepId][this.progressType];
			verticesInLUT.splice(verticesInLUT.indexOf(this),1);
		}
	},

	bindInvoke : function() {
		switch (this.type) {
			case "start"	: this.invoke = this.parentNode.start	; break;
			case "end"		: this.invoke = this.parentNode.end		; break;
			default			: this.invoke = this._nothing			; break;
		}
	},

	// Private Methods
	_nothing: function() {;}
}

//
// Transaction Module Initialization
//
var stepsLUT = {};
var transDoc = document.implementation.createDocument(null, null, null);

// this flag is used in all the browser types
transDoc.isXml = true;

Transaction._transactions = transDoc.appendChild(transDoc.createElement("transactions"));

// If not IE
if (Transaction.prototype.__proto__ ) {
	Transaction.prototype.__proto__ = transDoc.createElement("transaction").__proto__;
	Vertex.prototype.__proto__ = transDoc.createElement("vertex").__proto__;
}
//
// Transaction DOM Events Generators
//
// IE - the DOM NodeInserted is not working on other elements either then the own element
/*

transDoc.addEventListener("DOMNodeInserted", function(e) {
	if (e.target.tagName == "transaction" || e.target.tagName == "vertex")
		Event(e.target.tagName == "transaction" ? "transaction/nodeInserted" : "transaction/vertex/nodeInserted",
			e.target, { origEvent : e }).dispatch();
	}, false);

transDoc.addEventListener("DOMNodeRemoved", function(e) {
	if (e.target.tagName == "transaction" || e.target.tagName == "vertex")
		Event(e.target.tagName == "transaction" ? "transaction/nodeRemoved" : "transaction/vertex/nodeRemoved",
			e.target, { origEvent : e }).dispatch();
	}, false);

transDoc.addEventListener("DOMAttrModified", function(e) {
	if (e.target.tagName == "transaction" || e.target.tagName == "vertex")
		Event(e.target.tagName == "transaction" ? "transaction/attrModified" : "transaction/vertex/attrModified",
			e.target, { origEvent : e }).dispatch();
	}, false);
*/
//
// Transaction Related Rules
Query.addEventsListeners({
	// RULE: Invoke Vertices During Replay
	"step/progress" : function(e) {
		// Ignore the progress when the step failed
		if (typeof e.retCode != "undefined" && e.retCode > Step.RETCODE_SUCCEEDED)
			return;

		var LUT = stepsLUT[this.stepId];
		var verticesToInvoke = LUT && LUT[e.progressType];
		if (verticesToInvoke)
			for (var i = 0; i < verticesToInvoke.length; i++) {
				verticesToInvoke[i].parentNode.status = 0;
				verticesToInvoke[i].invoke.call(verticesToInvoke[i].parentNode);
			}
	},

	// RULE: When a vertex is created, associate an invoke function
	"transaction/vertex/nodeInserted" : function(e) { e.target.bindInvoke(); },

	// RULE: rebuild stepsLUT
	"transaction/nodeInserted transaction/vertex/nodeInserted transaction/nodeRemoved transaction/vertex/nodeRemoved transaction/attrModified transaction/vertex/attrModified" : function(e) {
		stepsLUT = {};
		var vertexes = find("xpath://vertex", Transaction._transactions);
		for (var i = 0; i < vertexes.length; i++) vertexes[i].insertToStepsLUT();
	},

	"loader/exposeRRE" :  function(e) {
		e.scope.Transaction = Transaction;
	},

	"RRE/getTransactions" : function(e) {
		e.transactions = Transaction._transactions;
	}
});

})();
